"""
CORE DATA MODELS
================
Contains the core data structure models for sequencing files.
"""

import logging
import os
import re
from .config import CONFIG as cfg

try:
	FileNotFoundError
except NameError:
	FileNotFoundError = OSError


logger = logging.getLogger(__name__)


def extract_frame(name):
	"""
	This function by default extracts the last set of digits in the
	string name and assumes it is the frame number when returning the
	parts. It's a good idea to only pass basenames without extensions
	so it doesn't attempt to sequence directory names or digits in the
	extension.

	:param str name: File basename without dir or extension.
	:return: 3-pair tuple consisting of the head (all characters
	         preceding the last set of digits), the frame number
	         (last set of digits), and tail (all digits succeeding
	         the frame number).
	"""
	frame_match = re.match(cfg.frame_extract_re, name)
	if frame_match:
		groups = frame_match.groups()
		head, tail = groups[cfg.head_group], groups[cfg.tail_group]
		frame = groups[cfg.frame_group]
	else:
		head, frame, tail = (name, '', '')
	if head is None:
		head = ''
	return head, frame, tail


def split_extension(filename):
	"""
	Splits the extension off the filename and returns a tuple of the
	base filename and the extension (without the dot).
	
	:param str filename: The base filename string to split.
	:return: A tuple of the head (characters before the last '.') and
			 the extension (characters after the last '.').
	"""
	parts = filename.split('.')
	if len(parts) < 2:
		return parts[0], ''
	ext = parts.pop(-1)
	head = '.'.join(parts)
	return head, ext


def frame_ranges_to_string(frames):
	"""
	Take a list of numbers and make a string representation of the ranges.

	>>> frame_ranges_to_string([1, 2, 3, 6, 7, 8, 9, 13, 15])
	'[1-3, 6-9, 13, 15]'

	:param list frames: Sorted list of frame numbers.
	:return: String of broken frame ranges (i.e '[10-14, 16, 20-25]').
	"""
	if not frames:
		return '[]'
	if not isinstance(frames, list):
		frames = list(frames)
	frames.sort()

	# Make list of lists for each consecutive range
	ranges = [[frames.pop(0)]]
	current_range_index = 0  # index of current range
	for x in frames:
		if x - 1 == ranges[current_range_index][-1]:
			ranges[current_range_index].append(x)
		else:
			current_range_index += 1
			ranges.append([x])
	range_strings = []
	for x in ranges:
		if len(x) > 1:
			range_strings.append('-'.join([str(x[0]), str(x[-1])]))
		else:
			range_strings.append(str(x[0]))
	complete_string = '[' + ', '.join(range_strings) + ']'
	return complete_string


class Stat(object):
	"""
	This class mocks objects returned by os.stat on Unix platforms.
	This is useful for instance when working with offline lists where 
	you want to maintain the stat information from a previously parsed
	directory which the current machine does not have access to.
	"""

	def __init__(self, size=None, ino=None, ctime=None, mtime=None,
				 atime=None, mode=None, dev=None, nlink=None, uid=None,
				 gid=None):
		"""
		Refer to the docs for the the built-in os.stat module for more info.
		
		:param int size: File size in bytes.
		:param int ino: Inode number.
		:param float ctime: Unix change timestamp.
		:param float mtime: Unix modify timestamp.
		:param float atime: Unix access timestamp.
		:param int mode: Inode protection mode.
		:param int dev: Device inode resides on.
		:param int nlink: Number of links to the inode.
		:param int uid: User id of the owner.
		:param int gid: Group id of the owner.
		"""
		self.st_size = size
		self.st_ino = ino
		self.st_nlink = nlink
		self.st_dev = dev
		self.st_mode = mode
		self.st_uid = uid
		self.st_gid = gid
		self.st_ctime = ctime
		self.st_mtime = mtime
		self.st_atime = atime

	def __getattr__(self, item):
		ints = ['st_size', 'st_ino', 'st_nlink', 'st_dev', 'st_mode',
				'st_uid', 'st_gid']
		floats = ['st_ctime', 'st_mtime', 'st_atime']
		try:
			if item in ints:
				return int(super(Stat, self).__getattribute__(item))
			elif item in floats:
				return float(super(Stat, self).__getattribute__(item))
		except TypeError:
			return None


class File(object):
	"""
	The class that represents a single file and all of it's Stat
	attributes if available. It contains attributes for the 
	various string parts of the path, base filename, frame number
	and extension. All Sequences are comprised of File objects.
	"""

	def __init__(self, filepath, stats=None, get_stats=None):
		"""
		Initalize a single File instance.

		:param str filepath: the absolute filepath of the file
		:param stats: dict or iterable to map Stat class or 
		              os.stat_result object.
		:param bool get_stats: True to attempt to call os.stat on
		                       the file. If file does not exist,
		                       revert back to applying stats values
		                       if they were supplied, else set stats
		                       to None.
		"""
		if get_stats is not None and isinstance(get_stats, bool):
			cfg.get_stats = get_stats
		self.abspath = filepath
		self.path, self.name = os.path.split(filepath)
		self._base, self.ext = split_extension(self.name)

		parts = extract_frame(self._base)
		self.namehead, self._framenum, tail = parts
		self.head = os.path.join(self.path, self.namehead)
		if not self.ext:
			self.tail = ''
		else:
			self.tail = '.'.join([tail, self.ext])
		self.padding = len(self._framenum)

		try:
			if get_stats:
				try:
					stats = os.stat(filepath)
				except FileNotFoundError:
					if stats is None:
						raise TypeError
			if isinstance(stats, os.stat_result):
				self.stat = stats
			elif isinstance(stats, dict):
				self.stat = Stat(**stats)
			elif isinstance(stats, (list, tuple)):
				self.stat = Stat(*stats)
			else:
				raise TypeError
		except TypeError:
			self.stat = Stat()

	def __str__(self):
		return self.abspath

	def __repr__(self):
		return "File('%s')" % self.abspath

	def __lt__(self, other):
		if isinstance(other, File):
			return self.frame < other.frame
		else:
			raise TypeError('%s not File instance.' % str(other))

	def __gt__(self, other):
		if isinstance(other, File):
			return self.frame > other.frame
		else:
			raise TypeError('%s not File instance.' % str(other))

	def __le__(self, other):
		if isinstance(other, File):
			return self.frame <= other.frame
		else:
			raise TypeError('%s not File instance.' % str(other))

	def __ge__(self, other):
		if isinstance(other, File):
			return self.frame >= other.frame
		else:
			raise TypeError('%s not File instance.' % str(other))

	def __eq__(self, other):
		if isinstance(other, File):
			return (self.head, self.frame, self.tail) == (other.head,
														  other.frame,
														  other.tail)
		else:
			return False

	def __ne__(self, other):
		if isinstance(other, File):
			return (self.head, self.frame, self.tail) != (other.head,
														  other.frame,
														  other.tail)
		else:
			return True

	@property
	def frame(self):
		""" Integer frame number. """
		try:
			return int(self._framenum)
		except ValueError:
			return None

	@property
	def frame_as_str(self):
		""" Str frame number with padding matching original filename. """
		return self._framenum

	@property
	def size(self):
		""" Same as Stat.st_size if available, otherwise None. """
		if not self.stat.st_size:
			try:
				self.stat.st_size = os.stat(self.abspath).st_size
				return self.stat.st_size
			except FileNotFoundError:
				return
		else:
			return self.stat.st_size

	@property
	def inode(self):
		""" Same as Stat.st_ino if available, otherwise None. """
		if not self.stat.st_ino:
			try:
				self.stat.st_ino = os.stat(self.abspath).st_ino
				return self.stat.st_ino
			except FileNotFoundError:
				return
		else:
			return self.stat.st_ino

	@property
	def nlink(self):
		""" Same as Stat.st_nlink if available, otherwise None. """
		if not self.stat.st_nlink:
			try:
				self.stat.st_nlink = os.stat(self.abspath).st_nlink
				return self.stat.st_nlink
			except FileNotFoundError:
				return
		else:
			return self.stat.st_nlink

	@property
	def dev(self):
		""" Same as Stat.st_dev if available, otherwise None. """
		if not self.stat.st_dev:
			try:
				self.stat.st_dev = os.stat(self.abspath).st_dev
				return self.stat.st_dev
			except FileNotFoundError:
				return
		else:
			return self.stat.st_dev

	@property
	def mode(self):
		""" Same as Stat.st_mode if available, otherwise None. """
		if not self.stat.st_mode:
			try:
				self.stat.st_mode = os.stat(self.abspath).st_mode
				return self.stat.st_mode
			except FileNotFoundError:
				return
		else:
			return self.stat.st_mode

	@property
	def uid(self):
		""" Same as Stat.st_uid if available, otherwise None. """
		if not self.stat.st_uid:
			try:
				self.stat.st_uid = os.stat(self.abspath).st_uid
				return self.stat.st_gid
			except FileNotFoundError:
				return
		else:
			return self.stat.st_uid

	@property
	def gid(self):
		""" Same as Stat.st_gid if available, otherwise None. """
		if not self.stat.st_gid:
			try:
				self.stat.st_gid = os.stat(self.abspath).st_gid
				return self.stat.st_gid
			except FileNotFoundError:
				return
		else:
			return self.stat.st_gid

	@property
	def ctime(self):
		""" Same as Stat.st_ctime if available, otherwise None. """
		if not self.stat.st_ctime:
			try:
				self.stat.st_ctime = os.stat(self.abspath).st_ctime
				return self.stat.st_ctime
			except FileNotFoundError:
				return
		else:
			return self.stat.st_ctime

	@property
	def mtime(self):
		""" Same as Stat.st_mtime if available, otherwise None. """
		if not self.stat.st_mtime:
			try:
				self.stat.st_mtime = os.stat(self.abspath).st_mtime
				return self.stat.st_mtime
			except FileNotFoundError:
				return
		else:
			return self.stat.st_mtime

	@property
	def atime(self):
		""" Same as Stat.st_atime if available, otherwise None. """
		if not self.stat.st_atime:
			try:
				self.stat.st_atime = os.stat(self.abspath).st_atime
				return self.stat.st_atime
			except FileNotFoundError:
				return
		else:
			return self.stat.st_atime

	def get_seq_key(self, ignore_padding=None):
		"""
		Make a sequence global name for matching frames to correct
		sequences.

		:param bool ignore_padding: True uses '%0#d' format, else '#'
		:return: sequence key name (i.e. '/path/to/file.#.dpx')
		"""
		if ignore_padding is None or not isinstance(ignore_padding, bool):
			ignore_padding = cfg.ignore_padding
		if not self._framenum:
			digits = ''
		elif ignore_padding is True:
			digits = '#'
		elif ignore_padding is False:
			digits = '%%0%dd' % self.padding
		else:
			raise TypeError('ignore_padding argument must be of type bool.')
		return self.head + digits + self.tail


class Sequence(object):
	"""
	Class representing a sequence of matching file names. The frames
	are stored in a dictionary with the frame numbers as keys. Sets
	are used for fast operations in calculating missing frames.

	This class's usage of dictionaries and sets is the core of the
	speed of this program. Rather than recursively searching existing
	sequences, the file key gnerated from File.get_seq_key() is used to
	instantly match the sequence it belongs to.
	"""

	def __init__(self, frame_file=None, ignore_padding=None):
		"""
		Initialize an image Sequence instance.

		:param File or str frame_file: File object or filename string to
		                         initialze a File object from.
		:param bool ignore_padding: True to allow inconsistent padding
		                            as same sequence, False to treat
		                            as separate sequences.
		"""
		if ignore_padding is not None and isinstance(ignore_padding, bool):
			cfg.ignore_padding = ignore_padding
		self._frames = {}
		self.seq_name = ''
		self.path = ''
		self.namehead = ''
		self.head = ''
		self.tail = ''
		self.ext = ''
		self.padding = 0
		self.inconsistent_padding = False
		if frame_file is not None:
			self.append(frame_file)

	def __str__(self):
		return self.format(cfg.format)

	def __repr__(self):
		return "Sequence('%s', frames=%d)" % (
			self.format(cfg.format), self.frames)

	def __len__(self):
		return len(self._frames)

	def __iter__(self):
		return iter([self._frames[frame] for frame in self._frames])

	def __getitem__(self, frames):
		all_frames = list(sorted(self._frames))
		if isinstance(frames, slice):
			return [self._frames[f] for f in sorted(self._frames)][frames]
		return self._frames[all_frames[frames]]

	def __lt__(self, other):
		if isinstance(other, str):
			return self.seq_name < other
		else:
			return self.seq_name < other.seq_name

	@property
	def abspath(self):
		""" Full sequence path name (i.e. '/path/to/file.#.dpx'). """
		return self.seq_name

	@property
	def name(self):
		""" The base sequence name without the path (i.e 'file.#.dpx'). """
		return os.path.basename(self.seq_name)

	@property
	def start(self):
		""" Int of first frame in sequence. """
		return min(self._frames)

	@property
	def end(self):
		""" Int of last frame in sequence. """
		return max(self._frames)

	@property
	def frames(self):
		""" Int of total frames in sequence. """
		return len(self)

	@property
	def frame_numbers(self):
		""" List of frame ints in sequence. """
		return list(sorted(self._frames))

	@property
	def frame_range(self):
		""" Int of expected frames in sequence. """
		return self.end - self.start + 1

	@property
	def missing(self):
		""" Int of total missing frames in sequence. """
		return self.end - (self.start - 1) - self.frames

	@property
	def is_missing_frames(self):
		""" Return True if any frames are missing from sequence. """
		return self.frames != self.frame_range

	@property
	def size(self):
		""" Sum of all filesizes (in bytes) in sequence. """
		try:
			return sum([file_.size for file_ in self])
		except TypeError:
			return

	def get_frame(self, frame):
		"""
		Get a specific frame number's File object, works like
		__getitem__ except gets exact frame number instead of index
		position in list.
		
		:param int frame: Frame number to extract from sequence.
		:return: File instance.
		"""
		return self._frames[int(frame)]

	def get_frames(self, start=None, end=None, step=1):
		"""
		Get a specific range of frames' File objects, works like
		__getitem__ slice, but gets exact frames or frame ranges
		instead of index positions in list.
		
		:param int start: Frame number to start range.
		:param int end: Frame number to end range.
		:param int step: Steps, like range function uses.
		:return: A list of File objects from the given ranges.
		"""
		frame_list = []
		if start is None:
			return self[:]
		elif end is not None:
			for frame in range(start, end + 1, step):
				if frame in self._frames:
					frame_list.append(self._frames[frame])
				else:
					logger.warning('Cannot get frame %d not in sequence %s.' %
								   (frame, str(self)))
		return frame_list

	def get_missing_frames(self):
		""" Get list of frame numbers missing in sequence. """
		implied = range(self.start, self.end + 1)
		return [frame for frame in implied if frame not in self._frames]

	def append(self, frame_file):
		"""
		Add a new frame to the sequence if it is has a matching
		sequence key. If it is the first file being added to a new
		Sequence object, it will set all the Sequence attributes.

		:param frame_file: File instance or string to append to Sequence.
		"""
		if not isinstance(frame_file, File):
			if isinstance(frame_file, str):
				frame_file = File(frame_file)
				if len(self._frames) > 0 and frame_file.get_seq_key(
						cfg.ignore_padding) != self.seq_name:
					raise ValueError('%s is not a member of %s. Not appending.'
					                 % (frame_file, repr(self)))
		if frame_file.frame is None:
			raise ValueError('%s can not be sequenced.' % str(frame_file))
		if not self.frames:
			self.namehead = frame_file.namehead
			self.path = frame_file.path
			self.head = frame_file.head
			self.tail = frame_file.tail
			self.ext = frame_file.ext
			self.padding = frame_file.padding
			self.seq_name = frame_file.get_seq_key(cfg.ignore_padding)
		elif frame_file.frame in self._frames:
			raise IndexError(
				'%s already in sequence as %s' % (
					frame_file.name, self._frames[frame_file.frame]))
		elif self.padding < frame_file.padding:
			self.inconsistent_padding = True
			self.padding = frame_file.padding
		self._frames[frame_file.frame] = frame_file

	def format(self, str_format=cfg.format):
		"""
		This formatter will replace any of the formatting directives
		found in the format argument with it's string part. It will try
		to format any character after a % sign, so in order to use a
		literal %, it must be escaped with another % - '%%'.

		+-------------------------------------------------------------------+
		| SAMPLE NAME:   '/path/to/file_name.0101.final.ext'                |
		+-------------------------------------------------------------------+

		+------+---------------------------------+--------------------------+
		| FMT  |  DESCRIPTION                    | EXAMPLE                  |
		+------+---------------------------------+--------------------------+
		| '%%' |  literal '%' sign               |  '%'                     |
		+------+---------------------------------+--------------------------+
		| '%p' |  pathname                       |  '/path/to'              |
		+------+---------------------------------+--------------------------+
		| '%h' |  head chars of filename         |  'file_name.'            |
		+------+---------------------------------+--------------------------+
		| '%H' |  all chars preceeding frame #   |  '/path/to/file_name.'   |
		+------+---------------------------------+--------------------------+
		| '%f' |  number of actual frames        |  '42'                    |
		+------+---------------------------------+--------------------------+
		| '%r' |  implied frame range            |  '[0101-0150]'           |
		+------+---------------------------------+--------------------------+
		| '%R' |  broken explicit frame range    |  '[101-140, 148, 150]'   |
		|      |  ignores padding                |                          |
		+------+---------------------------------+--------------------------+
		| '%m' |  total number of missing frames |  '8'                     |
		+------+---------------------------------+--------------------------+
		| '%M' |  broken explicit missing ranges |  '[141-147, 149]'        |
		|      |  ignores padding                |                          |
		+------+---------------------------------+--------------------------+
		| '%D' |  '#' digits denoting padding    |  '####'                  |
		+------+---------------------------------+--------------------------+
		| '%P' |  '%' style padding              |  '%04d'                  |
		+------+---------------------------------+--------------------------+
		| '%t' |  tail chars after frame, no ext |  '.final'                |
		+------+---------------------------------+--------------------------+
		| '%T' |  all tail chars after frame     |  '.final.ext'            |
		+------+---------------------------------+--------------------------+
		| '%e' |  extension without dot          |  'ext'                   |
		+------+---------------------------------+--------------------------+

		:param str_format: The string directive for the formatter to convert.
		:return: The formatted sequence string.
		"""

		# Call functions to minimize processes run during formatter execution.
		directive_mapper = {
			'%%': self.__pct,
			'%p': self.__path,
			'%h': self.__namehead,
			'%H': self.__head,
			'%f': self.__num_frames,
			'%r': self.__implied_range,
			'%R': self.__explicit_range,
			'%m': self.__num_missing_frames,
			'%M': self.__explicit_missing_range,
			'%D': self.__digits_pound_signs,
			'%P': self.__digits_padding,
			'%t': self.__tail_without_ext,
			'%T': self.__tail,
			'%e': self.__ext,
			}
		str_format = [c for c in str_format]
		new_string = ''

		matched = False
		for char in str_format:
			if matched:
				new_string += directive_mapper['%' + char]()
				matched = False
				continue
			if char == '%':
				matched = True
				continue
			else:
				new_string += char
		return new_string

	@staticmethod
	def __pct():
		""" Internal formatter method """
		return '%'

	def __path(self):
		""" Internal formatter method """
		return self.path

	def __namehead(self):
		""" Internal formatter method """
		return self.namehead

	def __head(self):
		""" Internal formatter method """
		return self.head

	def __num_frames(self):
		""" Internal formatter method """
		return str(self.frames)

	def __implied_range(self):
		""" Internal formatter method """
		return '[' + str(self.get_frame(self.start).frame_as_str) + \
			   '-' + str(self.get_frame(self.end).frame_as_str) + ']'

	def __explicit_range(self):
		""" Internal formatter method """
		return frame_ranges_to_string(sorted(self._frames))

	def __num_missing_frames(self):
		""" Internal formatter method """
		return str(self.missing)

	def __explicit_missing_range(self):
		""" Internal formatter method """
		return frame_ranges_to_string(self.get_missing_frames())

	def __digits_pound_signs(self):
		""" Internal formatter method """
		return '#' * self.padding

	def __digits_padding(self):
		""" Internal formatter method """
		return '%%0%dd' % self.padding

	def __tail_without_ext(self):
		""" Internal formatter method """
		return '.'.join(self.tail.split('.')[:-1])

	def __tail(self):
		""" Internal formatter method """
		return self.tail

	def __ext(self):
		""" Internal formatter method """
		return self.ext



"""
CONFIGURATION
=============
The configuration file setup module for ultrasequence. The cfg variable
contains the initialized Config instance which has either the default
attributes, or values found in an .ultrasequence.conf file if one was
available.

GLOBALS
-------
CONFIG

"""
import logging
try:
	import configparser
	PYTHON_VERSION = 3
except ImportError:
	import ConfigParser as configparser
	PYTHON_VERSION = 2
import os

logger = logging.getLogger()
logger.setLevel(logging.INFO)
logging.basicConfig()


class UsConfig(object):
	"""
	This class sets up a default configuration, and then tries to overload
	all the attributes with the values from a user configuration file. All
	available config options are accessible as instance attributes.
	"""
	def __init__(self):
		"""
		Init the default parser, and then overload with values found in a
		local config file.
		"""
		self.default_config = {
			'global': {
				'format': '%H%r%T',
				'recurse': 'false',
				'ignore_padding': 'true',
				'include_exts': '',
				'exclude_exts': '',
				'get_stats': 'false',
			},
			'regex': {
				'frame_extract': r'((.*)(\D))?(\d+)(.*)',
				'head_group': 0,
				'frame_group': 3,
				'tail_group': 4,
			}
		}
		self.user_config_file = os.path.expanduser('~/.ultrasequence.conf')
		self.default_parser = configparser.RawConfigParser()

		if PYTHON_VERSION == 3:
			self.default_parser.read_dict(self.default_config)
		else:
			self.default_parser.add_section('global')
			for key in self.default_config['global'].keys():
				self.default_parser.set(
					'global', key, self.default_config['global'][key])
			self.default_parser.add_section('regex')
			for key in self.default_config['regex'].keys():
				self.default_parser.set(
					'regex', key, self.default_config['regex'][key])

		self._load_config(self.default_parser)
		self._load_user_config()

	def __repr__(self):
		return (
			'Config({recurse={0}, ignore_padding={1}, include_exts={2},'
			'exclude_exts={3}, get_stats={4}, format={5})'.format(
				self.recurse, self.ignore_padding, self.include_exts,
				self.exclude_exts, self.get_stats, self.format))

	def _load_config(self, cfgparser):
		"""
		Assign all config values to Config instance attributes.

		:param cfgparser: the ConfigParser object to load values from
		"""
		self.format = cfgparser.get('global', 'format')
		self.recurse = cfgparser.getboolean('global', 'recurse')
		self.ignore_padding = cfgparser.getboolean('global', 'ignore_padding')
		self.include_exts = cfgparser.get('global', 'include_exts').split()
		self.exclude_exts = cfgparser.get('global', 'exclude_exts').split()
		self.get_stats = cfgparser.getboolean('global', 'get_stats')
		self.frame_extract_re = cfgparser.get('regex', 'frame_extract')
		self.head_group = cfgparser.getint('regex', 'head_group')
		self.frame_group = cfgparser.getint('regex', 'frame_group')
		self.tail_group = cfgparser.getint('regex', 'tail_group')

	def _load_user_config(self):
		""" Check for user config file and overload instance attributes. """
		if os.path.exists(self.user_config_file):
			cfgparser = configparser.RawConfigParser()
			cfgparser.read(self.user_config_file)
			self._load_config(cfgparser)

	def reset_defaults(self):
		self._load_config(self.default_parser)

	def write_user_config(self):
		""" Save a user config file with the default values. """
		with open(self.user_config_file, 'w') as f:
			self.default_parser.write(f)
		print('Made user config file at %s' % CONFIG.user_config_file)


CONFIG = UsConfig()



# -*- coding: utf-8 -*-

__version__ = '1.0.4'
NAME = 'ultrasequence'
DESCRIPTION = 'An extremely fast file sequencing tool.'
AUTHOR = 'Cody Cuellar'
EMAIL = 'cody.cuellar@gmail.com'
LICENSE = 'MIT'
COPYRIGHT = '©2018 Cody Cuellar'
URL = 'http://github.com/codycuellar/ultrasequence.git'



import os
import unittest
from unittest import TestCase
from ultrasequence import models
from ultrasequence.config import CONFIG


class TestExtractFrame(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()

	def test_frame_name(self):
		result = models.extract_frame('/path/to/file.1000.more.ext')
		self.assertTupleEqual(result, ('/path/to/file.', '1000', '.more.ext'))

	def test_number_only_name(self):
		result = models.extract_frame('1000.more.ext')
		self.assertTupleEqual(result, ('', '1000', '.more.ext'))

	def test_no_number_name(self):
		result = models.extract_frame('/path/to/file.more.ext')
		self.assertTupleEqual(result, ('/path/to/file.more.ext', '', ''))

	def test_number_only(self):
		result = models.extract_frame('1000')
		self.assertTupleEqual(result, ('', '1000', ''))

	def test_no_extension(self):
		result = models.extract_frame('/path/to/file.1000')
		self.assertTupleEqual(result, ('/path/to/file.', '1000', ''))

	def test_number_before_frame(self):
		result = models.extract_frame('/path/to/vid_v1_2018.10.exr')
		self.assertTupleEqual(result, ('/path/to/vid_v1_2018.', '10', '.exr'))


class TestSplitExtension(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()

	def test_split_single_dot(self):
		split = models.split_extension('test.ext')
		self.assertTupleEqual(split, ('test', 'ext'))

	def test_split_multiple_dots(self):
		split = models.split_extension('test.some.123.ext1')
		self.assertTupleEqual(split, ('test.some.123', 'ext1'))

	def test_split_no_dots(self):
		split = models.split_extension('testext1')
		self.assertTupleEqual(split, ('testext1', ''))


class TestFrameRangesToString(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()

	def test_convert_list(self):
		frames = [100, 101, 102, 104, 107, 108, 1010]
		result = models.frame_ranges_to_string(frames)
		self.assertEqual(result, '[100-102, 104, 107-108, 1010]')

	def test_convert_tuple(self):
		frames = (100, 101, 102, 104, 107, 108, 1010)
		result = models.frame_ranges_to_string(frames)
		self.assertEqual(result, '[100-102, 104, 107-108, 1010]')

	def test_convert_empty_list(self):
		result = models.frame_ranges_to_string([])
		self.assertEqual(result, '[]')

	def test_convert_single_item(self):
		result = models.frame_ranges_to_string([5])
		self.assertEqual(result, '[5]')


class TestStat(TestCase):
	def setUp(self):
		self.stat = models.Stat(1, 2, 3.1, 4.1, 5.1, 6, 7, 8, 9, 10)

	def test_st_size(self):
		self.assertIsInstance(self.stat.st_size, int)
		self.assertEqual(self.stat.st_size, 1)

	def test_st_ino(self):
		self.assertIsInstance(self.stat.st_ino, int)
		self.assertEqual(self.stat.st_ino, 2)

	def test_st_ctime(self):
		self.assertIsInstance(self.stat.st_ctime, float)
		self.assertEqual(self.stat.st_ctime, 3.1)

	def test_st_mtime(self):
		self.assertIsInstance(self.stat.st_mtime, float)
		self.assertEqual(self.stat.st_mtime, 4.1)

	def test_st_atime(self):
		self.assertIsInstance(self.stat.st_atime, float)
		self.assertEqual(self.stat.st_atime, 5.1)

	def test_st_mode(self):
		self.assertIsInstance(self.stat.st_mode, int)
		self.assertEqual(self.stat.st_mode, 6)

	def test_st_dev(self):
		self.assertIsInstance(self.stat.st_dev, int)
		self.assertEqual(self.stat.st_dev, 7)

	def test_st_nlink(self):
		self.assertIsInstance(self.stat.st_nlink, int)
		self.assertEqual(self.stat.st_nlink, 8)

	def test_st_uid(self):
		self.assertIsInstance(self.stat.st_uid, int)
		self.assertEqual(self.stat.st_uid, 9)

	def test_st_gid(self):
		self.assertIsInstance(self.stat.st_gid, int)
		self.assertEqual(self.stat.st_gid, 10)


class TestFile(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()
		self.file_10 = models.File('/some/file.10.dpx')
		self.file_11 = models.File('/some/file.11.dpx')
		self.file_011 = models.File('/some/file.011.dpx')
		self.file_012 = models.File('/some/file.012.dpx')

	def test_normal_init(self):
		_file = models.File('/path/to/file.01000.more.ext')
		self.assertEqual(_file.abspath, '/path/to/file.01000.more.ext')
		self.assertEqual(_file.path, '/path/to')
		self.assertEqual(_file.name, 'file.01000.more.ext')
		self.assertEqual(_file.ext, 'ext')
		self.assertEqual(_file.namehead, 'file.')
		self.assertEqual(_file.head, '/path/to/file.')
		self.assertEqual(_file.frame_as_str, '01000')
		self.assertEqual(_file.frame, 1000)
		self.assertEqual(_file.tail, '.more.ext')
		self.assertEqual(_file.padding, 5)

	def test_no_number_init(self):
		_file = models.File('/path/to/file.ext')
		self.assertEqual(_file.abspath, '/path/to/file.ext')
		self.assertEqual(_file.path, '/path/to')
		self.assertEqual(_file.name, 'file.ext')
		self.assertEqual(_file.ext, 'ext')
		self.assertEqual(_file.namehead, 'file')
		self.assertEqual(_file.head, '/path/to/file')
		self.assertEqual(_file.frame_as_str, '')
		self.assertEqual(_file.frame, None)
		self.assertEqual(_file.tail, '.ext')
		self.assertEqual(_file.padding, 0)

	def test_number_only_init(self):
		_file = models.File('/path/to/01234.ext')
		self.assertEqual(_file.abspath, '/path/to/01234.ext')
		self.assertEqual(_file.path, '/path/to')
		self.assertEqual(_file.name, '01234.ext')
		self.assertEqual(_file.ext, 'ext')
		self.assertEqual(_file.head, '/path/to/')
		self.assertEqual(_file._framenum, '01234')
		self.assertEqual(_file.frame, 1234)
		self.assertEqual(_file.tail, '.ext')
		self.assertEqual(_file.padding, 5)

	def test_number_only_no_path_init(self):
		_file = models.File('1234.ext')
		self.assertEqual(_file.abspath, '1234.ext')
		self.assertEqual(_file.path, '')
		self.assertEqual(_file.name, '1234.ext')
		self.assertEqual(_file.ext, 'ext')
		self.assertEqual(_file.head, '')
		self.assertEqual(_file.frame_as_str, '1234')
		self.assertEqual(_file.frame, 1234)
		self.assertEqual(_file.tail, '.ext')
		self.assertEqual(_file.padding, 4)

	def test_use_stat_dict(self):
		stats = {
			'size': 1,
			'ino': 2,
			'nlink': 3,
			'dev': 4,
			'mode': 5,
			'uid': 6,
			'gid': 7,
			'mtime': 8,
			'ctime': 9,
			'atime': 10
		}
		_file = models.File('filename.ext', stats=stats)
		self.assertEqual(_file.size, 1)
		self.assertEqual(_file.inode, 2)
		self.assertEqual(_file.nlink, 3)
		self.assertEqual(_file.dev, 4)
		self.assertEqual(_file.mode, 5)
		self.assertEqual(_file.uid, 6)
		self.assertEqual(_file.gid, 7)
		self.assertEqual(_file.mtime, 8)
		self.assertEqual(_file.ctime, 9)
		self.assertEqual(_file.atime, 10)

	def test_use_stat_list(self):
		stats = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
		_file = models.File('filename.ext', stats=stats)
		self.assertIsInstance(_file.stat, models.Stat)
		self.assertEqual(_file.size, 1)
		self.assertEqual(_file.inode, 2)
		self.assertEqual(_file.ctime, 3)
		self.assertEqual(_file.mtime, 4)
		self.assertEqual(_file.atime, 5)
		self.assertEqual(_file.mode, 6)
		self.assertEqual(_file.dev, 7)
		self.assertEqual(_file.nlink, 8)
		self.assertEqual(_file.uid, 9)
		self.assertEqual(_file.gid, 10)

	def test_file_os_stat(self):
		_file = models.File(__file__, os.stat(__file__))
		self.assertIsInstance(_file.stat, os.stat_result)

	def test_get_stat(self):
		_file = models.File(__file__, get_stats=True)
		self.assertIsInstance(_file.stat, os.stat_result)

	def test_get_stats_unsuccessful_fallback(self):
		stats = {'size': 15}
		_file = models.File('/not/a/file/path.none', stats=stats,
						   get_stats=True)
		self.assertIsInstance(_file.stat, models.Stat)
		self.assertEqual(_file.size, 15)
		self.assertIsNone(_file.inode)

	def test_get_stats_override_supplied_stats(self):
		stats = {'size': -15, 'ino': None}
		_file = models.File(__file__, stats=stats, get_stats=True)
		self.assertIsInstance(_file.stat, os.stat_result)
		self.assertNotEqual(_file.size, -15)
		self.assertIsNotNone(_file.inode)

	def test_lt_same_padding(self):
		self.assertLess(self.file_10, self.file_11)

	def test_lt_different_padding(self):
		self.assertLess(self.file_11, self.file_012)

	def test_lt_not_equal(self):
		self.assertFalse(self.file_10 < self.file_10)

	def test_le_equal(self):
		self.assertLessEqual(self.file_10, self.file_10)

	def test_le_same_padding(self):
		self.assertLessEqual(self.file_10, self.file_11)

	def test_le_different_padding(self):
		self.assertLessEqual(self.file_11, self.file_012)

	def test_gt_same_padding(self):
		self.assertGreater(self.file_11, self.file_10)

	def test_gt_different_padding(self):
		self.assertGreater(self.file_012, self.file_11)

	def test_gt_not_equal(self):
		self.assertFalse(self.file_10 > self.file_10)

	def test_ge_equal(self):
		self.assertGreaterEqual(self.file_10, self.file_10)

	def test_ge_same_padding(self):
		self.assertGreaterEqual(self.file_11, self.file_10)

	def test_ge_different_padding(self):
		self.assertGreaterEqual(self.file_012, self.file_11)

	def test_eq_same_padding(self):
		self.assertEqual(self.file_11, self.file_11)

	def test_eq_different_padding(self):
		self.assertEqual(self.file_11, self.file_011)

	def test_eq_different_file_fail(self):
		other = models.File('/a/different/file.11.exr')
		self.assertFalse(self.file_11 == other)

	def test_ne_same_basename(self):
		self.assertNotEqual(self.file_10, self.file_11)

	def test_ne_different_basename_same_number(self):
		other = models.File('/a/different/file.11.exr')
		self.assertNotEqual(self.file_11, other)

	def test_get_seq_key_no_padding(self):
		_file = models.File('/path/to/file.1000.ext')
		self.assertEqual(_file.get_seq_key(True), '/path/to/file.#.ext')

	def test_get_seq_key_padding(self):
		_file = models.File('/path/to/file.1000.ext')
		self.assertEqual(_file.get_seq_key(ignore_padding=False),
						 '/path/to/file.%04d.ext')

	def test_get_seq_key_no_framenum(self):
		_file = models.File('/path/to/file.ext')
		self.assertEqual(_file.get_seq_key(True), '/path/to/file.ext')
		self.assertEqual(_file.get_seq_key(ignore_padding=False),
						 '/path/to/file.ext')


class TestSequence(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()
		missing_files = [
			'/abs/path/to/file_0100_name.ext',
			'/abs/path/to/file_0101_name.ext',
			'/abs/path/to/file_0102_name.ext',
			'/abs/path/to/file_0104_name.ext',
			'/abs/path/to/file_0107_name.ext',
			'/abs/path/to/file_0108_name.ext',
			'/abs/path/to/file_01010_name.ext',
		]
		self.missing = models.Sequence()
		for m in missing_files:
			self.missing.append(m)
		contiguous_files = [
			'/abs/path/to/file_0100_name.ext',
			'/abs/path/to/file_0101_name.ext',
			'/abs/path/to/file_0102_name.ext',
			'/abs/path/to/file_0103_name.ext',
		]
		self.contiguous = models.Sequence()
		for m in contiguous_files:
			self.contiguous.append(m)

	def test_sequence_init_append(self):
		seq = models.Sequence('/path/to/file.0100.ext')
		self.assertListEqual(list(seq._frames), [100])
		self.assertEqual(seq.seq_name, '/path/to/file.#.ext')
		self.assertEqual(seq.namehead, 'file.')
		self.assertEqual(seq.head, '/path/to/file.')
		self.assertEqual(seq.path, '/path/to')
		self.assertEqual(seq.tail, '.ext')
		self.assertEqual(seq.ext, 'ext')
		self.assertEqual(seq.padding, 4)
		self.assertEqual(seq.inconsistent_padding, False)

	def test_sequence_append_file(self):
		file1 = models.File('/path/to/file.0100.ext')
		seq = models.Sequence(file1)
		seq.append(models.File('/path/to/file.0101.ext'))
		self.assertEqual(seq.frames, 2)

	def test_sequence_append_string(self):
		_file = '/path/to/file.0100.ext'
		seq = models.Sequence(_file)
		seq.append('/path/to/file.0101.ext')
		self.assertEqual(seq.frames, 2)

	def test_sequence_append_exact_dupe(self):
		_file = models.File('/path/to/file.0100.ext')
		seq = models.Sequence(_file)
		with self.assertRaises(IndexError):
			seq.append('/path/to/file.0100.ext')

	def test_sequence_append_non_member(self):
		_file = models.File('/path/to/file.0100.ext')
		seq = models.Sequence(_file)
		with self.assertRaises(ValueError):
			seq.append('/not/a/member.0001.ext')

	def test_sequence_inconsistent_padding_dupe(self):
		_file = models.File('/path/to/file.0100.ext')
		seq = models.Sequence(_file)
		self.assertEqual(seq.seq_name, '/path/to/file.#.ext')
		with self.assertRaises(IndexError):
			seq.append('/path/to/file.00100.ext')

	def test_sequence_force_consistent_padding(self):
		_file = models.File('/path/to/file.0100.ext')
		seq = models.Sequence(_file, ignore_padding=False)
		self.assertEqual(seq.seq_name, '/path/to/file.%04d.ext')
		with self.assertRaises(ValueError):
			seq.append('/path/to/file.00100.ext')

	def test_sequence_size(self):
		files = [
			('file.0.ext', {'size': 10}),
			('file.1.ext', {'size': 10}),
			('file.2.ext', {'size': 10}),
		]
		sequence = models.Sequence()
		for _file in files:
			_file = models.File(*_file)
			sequence.append(_file)
		self.assertEqual(sequence.size, 30)

	def test_get_missing_frames(self):
		files = [
			'/abs/path/to/file_0100_name.ext',
			'/abs/path/to/file_0101_name.ext',
			'/abs/path/to/file_0103_name.ext',
			'/abs/path/to/file_0105_name.ext',
			'/abs/path/to/file_0106_name.ext',
			'/abs/path/to/file_0107_name.ext',
		]
		seq = models.Sequence()
		for _file in files:
			seq.append(_file)
		self.assertListEqual(seq.get_missing_frames(), [102, 104])


class TestSequenceFormatting(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()
		files = [
			'/abs/path/to/file_0100_name.ext',
			'/abs/path/to/file_0101_name.ext',
			'/abs/path/to/file_0102_name.ext',
			'/abs/path/to/file_0104_name.ext',
			'/abs/path/to/file_0107_name.ext',
			'/abs/path/to/file_0108_name.ext',
			'/abs/path/to/file_0110_name.ext',
		]
		self.seq = models.Sequence()
		for f in files:
			self.seq.append(f)

	def test_format_path(self):
		self.assertEqual(self.seq.format('%p'), '/abs/path/to')
		self.assertEqual(self.seq.format('test%ptest%p'),
						 'test/abs/path/totest/abs/path/to')

	def test_format_namehead(self):
		self.assertEqual(self.seq.format('%h'), 'file_')
		self.assertEqual(self.seq.format('test%htest%h'),
						 'testfile_testfile_')

	def test_format_head(self):
		self.assertEqual(self.seq.format('%H'), '/abs/path/to/file_')
		self.assertEqual(self.seq.format('test%Htest%H'),
						 'test/abs/path/to/file_test/abs/path/to/file_')

	def test_format_numframes(self):
		self.assertEqual(self.seq.format('%f'), '7')
		self.assertEqual(self.seq.format('test%ftest(%f)'),
						 'test7test(7)')

	def test_format_implied_range(self):
		self.assertEqual(self.seq.format('%r'), '[0100-0110]')
		self.assertEqual(self.seq.format('test%rtest%r'),
						 'test[0100-0110]test[0100-0110]')

	def test_format_explicit_range(self):
		self.assertEqual(self.seq.format('%R'),
						 '[100-102, 104, 107-108, 110]')
		self.assertEqual(
			self.seq.format('test%Rtest%R'),
			'test[100-102, 104, 107-108, 110]test[100-102, 104, 107-108, 110]')

	def test_format_num_missing_frames(self):
		self.assertEqual(self.seq.format('%m'), '4')
		self.assertEqual(self.seq.format('test%mtest%m'),
						 'test4test4')

	def test_format_explicit_missing_frames(self):
		self.assertEqual(self.seq.format('%M'), '[103, 105-106, 109]')
		self.assertEqual(self.seq.format('test%Mtest%M'),
						 'test[103, 105-106, 109]test[103, 105-106, 109]')

	def test_format_pound_padding(self):
		self.assertEqual(self.seq.format('%D'), '####')
		self.assertEqual(self.seq.format('test%Dtest%D'),
						 'test####test####')

	def test_format_formatted_padding(self):
		self.assertEqual(self.seq.format('%P'), '%04d')
		self.assertEqual(self.seq.format('test%Ptest%P'),
						 'test%04dtest%04d')

	def test_format_tail_without_ext(self):
		self.assertEqual(self.seq.format('%t'), '_name')
		self.assertEqual(self.seq.format('test%ttest%t'),
						 'test_nametest_name')

	def test_format_tail(self):
		self.assertEqual(self.seq.format('%T'), '_name.ext')
		self.assertEqual(self.seq.format('test%Ttest%T'),
						 'test_name.exttest_name.ext')

	def test_format_ext(self):
		self.assertEqual(self.seq.format('%e'), 'ext')
		self.assertEqual(self.seq.format('test%etest%e'),
						 'testexttestext')

	def test_format_pct(self):
		self.assertEqual(self.seq.format('%%'), '%')
		self.assertEqual(self.seq.format('100%% Success'),
						 '100% Success')

	def test_format_invalid_directive(self):
		with self.assertRaises(KeyError):
			self.seq.format('Try%^Fail')


if __name__ == '__main__':
	unittest.main()



"""
PARSING
=======
This module contains functions and classes for parsing files and directories
for file sequences.
"""
import logging
import os
import sys
from os import walk
from ultrasequence.config import CONFIG as cfg
from ultrasequence.models import File, Sequence


logger = logging.getLogger(__name__)


if sys.version_info < (3, 5):
	try:
		from scandir import walk
	except ImportError:
		logger.warning('For Python versions < 3.5, scandir module is '
		               'recommended for faster directory parsing. Run:'
		               '\n>>> pip install scandir')


def scan_dir(path):
	"""
	Searches a root directory and returns a list of all files. If 
	cfg.recurse is True, the scanner will descend all child directories.

	:param str path: The root path to scan for files.
	:return: A list of filenames if cfg.get_stats is False, or a list
			 of tuples (filename, file_stats) if cfg.get_stats is True.
	"""
	file_list = []
	if cfg.recurse:
		for root, dirs, files in walk(path):
			file_list += stat_files(root, files)
	else:
		file_list += stat_files(path, os.listdir(path))
	return file_list


def stat_files(root, files):
	"""
	Assembles a list of files for a single directory.

	:param str root: The the root path to the current directory.
	:param list files: The list of filenames in the directory.
	:return: a list of filenames if cfg.get_stats is False, or a list
			 of tuples (filename, file_stats) if cfg.get_stats is True.
	"""
	dir_list = []
	if cfg.get_stats:
		for file_ in files:
			abspath = os.path.join(root, file_)
			# TODO: For links, copy the stat from source, but set size to 0
			# if os.path.islink(abspath):
			# 	continue
			dir_list.append((abspath, os.stat(abspath)))
	else:
		dir_list += [os.path.join(root, file_) for file_ in files
					 if os.path.isfile(os.path.join(root, file_))]
	return dir_list


class Parser(object):
	"""
	The main parser class which handles all data parsing and sorting. It
	can parse a text file with file names, or it can scan a directory on
	a locally mounted volume. It handles include/exclude patterns, and
	controls weather to get stats from a file on disk or not. Once the
	parser has run, it will populate five important attributes:

		Parser.sequences
			A list of sequence objects with two or more frames.
		Parser.orphan_frames
			Any sequences of files that were only one frame long.
		Parser.no_frame_numbers
			Files that could not be parsed for frame numbers, thus could
			not be sequenced.
		Parser.excluded
			Files with extensions that are either not in the include list
			if one exists, or that are in the exclude list.
		Parser.collisions
			If ignore_padding is used when setting up the parser, it is
			possible that two files attempt to insert into the same sequence,
			for instance, file.001.dpx and file.01.dpx will be treated as
			frame 1 of sequence file.#.dpx, therefor the second frame that
			is found that already exists in the matching sequence will get
			added to collisions.

	>>> from ultrasequence import Parser
	>>> include = ['tif', 'jpg', 'dpx', 'exr']
	>>> parser = Parser(include_exts=include, ignore_padding=True)
	>>> parser.parse_directory('~/Desktop', recures=True)
	>>> print(parser)
	'Parser(sequences=2, orphan_frames=10, no_frame_numbers=5,
	excluded=10, collisions=0)'
	"""

	def __init__(self, include_exts=cfg.include_exts,
	             exclude_exts=cfg.exclude_exts, get_stats=cfg.get_stats,
	             ignore_padding=cfg.ignore_padding):
		"""
		Main parser class. Sets up config parameters for parsing methods.
		
		:param list include_exts: file extensions to include in parsing
		:param list exclude_exts: file extensions to explicitly exclude in
		                          parsing
		:param bool get_stats: get file stats from os.stats
		:param bool ignore_padding: ignore the number of digits in the
		                            file's frame number section
		"""
		if not include_exts or not isinstance(include_exts, (tuple, list)):
			self.include_exts = set()
		else:
			self.include_exts = set(include_exts)

		if not exclude_exts or not isinstance(exclude_exts, (tuple, list)):
			self.exclude_exts = set()
		else:
			self.exclude_exts = set(exclude_exts)

		cfg.get_stats = get_stats
		self.ignore_padding = ignore_padding
		self._reset()

	def _reset(self):
		""" Clear and init all parser results. """
		self._sequences = {}
		self.sequences = []
		self.orphan_frames = []
		self.no_frame_numbers = []
		self.excluded = []
		self.collisions = []
		self.parsed = False

	def __str__(self):
		return ('Parser(sequences=%d, orphan_frames=%d, no_frame_numbers=%d, '
				'excluded=%d, collisions=%d)' %
				(len(self.sequences), len(self.orphan_frames),
				 len(self.no_frame_numbers), len(self.excluded),
				 len(self.collisions)))

	def __repr__(self):
		return ('<Parser object at %s, parsed=%s>' %
				(hex(id(self)), self.parsed))

	def _cleanup(self):
		""" Moves single frames out of sequences list. """
		while self._sequences:
			seq = self._sequences.popitem()[1]
			if seq.frames == 1:
				self.orphan_frames.append(seq[0])
			else:
				self.sequences.append(seq)
		self.parsed = True

	def _sort_file(self, filepath, stats=None):
		""" Finds matching sequence for given filepath. """
		file_ = File(filepath, stats=stats)

		if self.include_exts and file_.ext.lower() not in self.include_exts \
				or file_.ext.lower() in self.exclude_exts:
			self.excluded.append(file_)

		elif file_.frame is None:
			self.no_frame_numbers.append(file_)

		else:
			seq_name = file_.get_seq_key()
			if seq_name in self._sequences:
				try:
					self._sequences[seq_name].append(file_)
				except IndexError:
					self.collisions.append(file_)
			else:
				self._sequences[seq_name] = Sequence(file_)

	def parse_directory(self, directory, recurse=cfg.recurse):
		"""
		Parse a directory on the file system.

		:param str directory: Directory path to scan on filesystem.
		:param bool recurse: Recurse all child directories.
		"""
		self._reset()
		cfg.recurse = recurse
		directory = os.path.expanduser(directory)
		if isinstance(directory, str) and os.path.isdir(directory):
			file_list = scan_dir(directory)
			while file_list:  # reduce memory consumption for large lists
				file_ = file_list.pop(0)
				if cfg.get_stats:
					self._sort_file(file_[0], file_[1])
				else:
					self._sort_file(file_)
			self._cleanup()
		else:
			logger.warning('%s is not an available directory.' % directory)

	def parse_file(self, input_file):
		"""
		Parse a text file containing file listings.

		:param str input_file: Path to the file containing a file listing.
		"""
		input_file = os.path.expanduser(input_file)

		self._reset()
		if isinstance(input_file, str) and os.path.isfile(input_file):
			with open(input_file, 'r') as file_list:
				for file_ in file_list:
					self._sort_file(file_.rstrip())
			self._cleanup()
		else:
			logger.warning('%s is not a valid filepath.' % input_file)



# -*- coding: utf-8 -*-
"""
ULTRASEQUENCE
=============
** An ultra-fast file sequencing tool. **

©2018 Cody Cuellar
cody.cuellar@gmail.com

License: MIT

--------------------------------------------------

This tool was built with speed in mind. It is very easy to use, 
highly customizable, and has many neat features to help with file
management.

The core function is sequencing files with matching basenames and
incremental 'frame' digits, which is very useful in film and VFX
workflows. This tool
makes it easy to sort them out and group them together and make it
easy to list seqeunces as single entries in a text manifest.

These files:
	/some/file.1000.dpx
	/some/file.1001.dpx
	/some/file.1002.dpx
	/some/file.1003.dpx
	/some/file.1004.dpx

Become:
	/some/file.[1000-1004].dpx

or by using the format method of the Sequence class such as:

>>> seq.format('%r %h%P%T Missing Frames: %m')
'[1000-1004] file.%04d.dpx Missing Frames: 0'

On top of providing simple formatting, it can parse directories, a
text file containing a list of file names, or even just a python list
of file path strings. It can be used both online or offline (whether
or not you have direct access to the file paths) and can fetch all os
stats from the file system if you do have access to the files.
"""
from ultrasequence.models import (
	extract_frame, split_extension, frame_ranges_to_string,
	Stat, File, Sequence)
from ultrasequence.parsing import scan_dir, stat_files, Parser
from ultrasequence.version import __version__

__all__ = [
	'extract_frame', 'split_extension', 'frame_ranges_to_string',
	'Stat', 'File', 'Sequence', 'scan_dir', 'stat_files', 'Parser'
	]



import unittest
import os
from unittest import TestCase
try:
	from unittest.mock import patch
except ImportError:
	from mock import patch
from ultrasequence import parsing
from ultrasequence.config import CONFIG


def stat_mock(root, files):
	dir_list = [os.path.join(root, file) for file in files]
	return dir_list


class TestScanDir(TestCase):
	def setUp(self):
		CONFIG.reset_defaults()
		self.walk = [
			(
				'/root',
				['seq_one', 'seq_two', 'seq_three'],
			 	['file_a.ext', 'file_b.ext']
			),
			(
				'/root/seq_one',
				[],
				['seq_one.1000.dxp', 'seq_one.1001.dxp', 'seq_one.1002.dxp']
			),
			(
				'/root/seq_two',
				['seq_two.0.dxp', 'seq_two.1.dxp', 'seq_two.3.dxp'],
				[]
			),
			(
				'/root/seq_three',
				[],
				['seq_three.mov']
			)
		]

	@patch('os.listdir')
	def test_scan_dir_default_no_recurse(self, mock_listdir):
		mock_listdir.return_value = self.walk[0][2]
		with patch('ultrasequence.parsing.stat_files') as mock_stat_files:
			mock_stat_files.side_effect = stat_mock
			result = parsing.scan_dir('/root')
			expected = [os.path.join('/root', file)
						for file in self.walk[0][2]]
			self.assertListEqual(result, expected)

	@patch('ultrasequence.parsing.walk')
	def test_scan_dir_recurse(self, mock_walk):
		CONFIG.recurse = True
		mock_walk.return_value = self.walk
		with patch('ultrasequence.parsing.stat_files') as mock_stat_files:
			mock_stat_files.side_effect = stat_mock
			result = parsing.scan_dir('/root')
			expected = []
			for root, dirs, files in self.walk:
				expected += [os.path.join(root, file) for file in files]
			self.assertListEqual(result, expected)

	@patch('os.path.isfile', return_value=True)
	def test_stat_files_default_no_stats(self, mock_isfile):
		result = parsing.stat_files('/root', self.walk[0][2])
		expected = [os.path.join('/root', file) for file in self.walk[0][2]]
		self.assertListEqual(result, expected)

	@patch('os.path.islink', return_value=False)
	def test_stat_files_enable_stats(self, mock_islink):
		CONFIG.get_stats = True
		with patch('os.stat', return_value='stats'):
			result = parsing.stat_files('/root', self.walk[0][2])
		expected = [(os.path.join('/root', file), 'stats')
					for file in self.walk[0][2]]
		self.assertListEqual(result, expected)


if __name__ == '__main__':
	unittest.main()



"""
Ultrasequence is a highly efficient Python image sequencer. It will parse
a list or directory of filenames and sequence any files that share a common
name with incremental frame numbers. It comes with a command line tool and is
highly customisable. It operates much faster than most other tools due to it's
use of key matching rather than recursively searching existing sequences for
matches. It's speed should be roughly O(n).
"""
from setuptools import setup, find_packages
import os


packagedir = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(packagedir, 'ultrasequence', 'version.py'), 'r') as v:
	exec(v.read(), globals())


setup(
	packages=find_packages(),
	name=NAME,
	version=__version__,
	author=AUTHOR,
	author_email=EMAIL,
	package_dir={'ultrasequece': 'ultrasequence'},
	url=URL,
	license=LICENSE,
	description=DESCRIPTION,
	long_description=open('readme.rst').read(),
	keywords='sequence file parser image ultra frames',
	platforms=['MacOS 10.10', 'MacOS 10.11', 'MacOS 10.12', 'MacOS 10.13'],
	python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
	install_requires=[],
	entry_points={
		'console_scripts': [
			'findseq=ultrasequence.bin.findseq:main'
			]
		}
	)



"""
FindSeq
=======
This module is the command-line utility entry-point. It parses the command-
line for arguments, and overrides any pre-loaded configs with those arguments.
The order of operation option loading is:

# Load default config defined in the ultrasequence.config.UsConfig class the
CONFIG instance
# Check for ~/.ultrasequence.conf and overload all parameters in the CONFIG
instance
# Parse command line and overwrite any options present to CONFIG

For instance, if no config file is present, the defaults will be loaded. If the
user provides the --format flag, the command line values will overwrite the
defaults in CONFIG. On the other hand, if a config file is present and the
user puts no options on the command line, all the values from the local config
file will be used.
"""
import argparse
import logging
import os
from ultrasequence.config import CONFIG as cfg
from ultrasequence.parsing import Parser
from ultrasequence.version import __version__, NAME


logger = logging.getLogger(__name__)


class UserConfig(argparse.Action):
	"""
	A custom class which allows an operation-and-exit style behaviour like
	with --help or --version, but for the case of writing a config file.
	"""
	def __init__(self,
				 option_strings,
				 dest=argparse.SUPPRESS,
				 default=argparse.SUPPRESS,
				 help_="make a local config file in user home dir and exit"):
		super(UserConfig, self).__init__(
			option_strings=option_strings,
			dest=dest,
			default=default,
			nargs=0,
			help=help_)

	def __call__(self, parser, namespace, values, option_string=None):
		cfg.write_user_config()
		parser.exit()


def get_args():
	""" Parse the command line args and return an argparser object. """
	parser = argparse.ArgumentParser(prog=NAME)

	parser.add_argument('source',
						type=str,
						help='path to source file or directory to scan.'
						)

	parser.add_argument('-v', '--version',
						action='version',
						version='%s v%s' % (NAME, __version__)
						)

	parser.add_argument('--make-config',
						action=UserConfig,
						)

	parser.add_argument('-I', '--ignore-config',
						action='store_true',
						help='ignore the user config file if it exists'
						)

	parser.add_argument('-f', '--format',
						type=str,
						help=''
						)

	parser.add_argument('-i', '--include',
						type=str,
						nargs='+',
						help='list of file extensions without preceding dot '
							 'to include in sequencer'
						)

	parser.add_argument('-e', '--exclude',
						type=str,
						nargs='+',
						help='list of file extensions without preceding dot '
							 'to exclude in sequencer'
						)

	parser.add_argument('-R', '--recurse',
						action='store_true',
						help='enable directory recursion for directory parser.'
						)

	parser.add_argument('-s', '--get-stats',
						action='store_true',
						help=''
						)

	parser.add_argument('-p', '--strict-padding',
						action='store_true',
						help='disables the ignore_padding rule. This will '
							 'treat all files with different amount of digit '
							 'padding as separate sequences.'
						)

	return parser.parse_args()


def main():
	""" Set up the args and run the parser. """
	args = get_args()

	if args.ignore_config:
		cfg.reset_defaults()
	if args.format:
		cfg.format = args.format
	if args.include:
		cfg.include_exts = args.include
	if args.exclude:
		cfg.exclude_exts = args.exclude
	if args.recurse:
		cfg.recurse = True
	if args.strict_padding:
		cfg.ignore_padding = False

	parser = Parser(include_exts=cfg.include_exts,
	                exclude_exts=cfg.exclude_exts,
	                get_stats=cfg.get_stats,
	                ignore_padding=cfg.ignore_padding)

	if os.path.isdir(args.source):
		parser.parse_directory(args.source, recurse=cfg.recurse)
	elif os.path.isfile(os.path.expanduser(args.source)):
		parser.parse_file(args.source)

	output = parser.sequences + parser.orphan_frames + \
	         parser.no_frame_numbers + parser.collisions
	output.sort(key=lambda s: s.abspath.lower())

	for x in output:
		print(x)


if __name__ == '__main__':
	main()
